# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Ephemerides functions for the sluyspy package"""


from astroquery.jplhorizons import Horizons as _Horizons


def horizons_ephem(obj, epochs, loc='500@399'):
    """Return a Pandas DataFrame with ephemerides from the NASA JPL Horizons system, with my selection of
    variables and column names.
    
    Parameters:
      obj (str):            Object using Horizons definitions e.g. 199-899 for the centres of the planets 1-8.
      epochs (float/dict):  List of JDs (<~50!) or start/stop/step dict for ephemeride epochs, e.g. Julian days:
                            start/stop/step dict: e.g. {'start':'2010-01-01', 'stop':'2010-01-10', 'step':'1d'}
      loc (str):            Horizons string for observer location.  Optional, defaults to '500@399' == Geocentre.
    
    Returns:
      (pd.df):  Pandas DataFrame containing planet ephemerides data.
    """
    
    hz = _Horizons(id=obj, location=loc, epochs=epochs)  # Return only quantities of interest
    
    eph = hz.ephemerides(quantities='30, 18,19, 31,20,39, 1,2,36, 43,10,11,9', extra_precision=True)  # 45 gives error?
    # optional_settings={'suppress_range_rate':'yes'})  # Documented, but gives error?
    # optional_settings: dict, optional: key-value based dictionary to inject some additional optional settings - see https://ssd.jpl.nasa.gov/horizons/app.html; default: empty
    
    # Convert astropy Table to Pandas DataFrame:
    df = eph.to_pandas()
    
    # Drop unwanted columns:
    df = df.drop(['targetname','solar_presence','flags','r_rate','delta_rate','r_3sigma','r_rate_3sigma',
                  'RA_3sigma','DEC_3sigma','PABLon','PABLat','surfbright'], axis=1)
    
    # Rename columns:
    df = df.rename(columns={'datetime_str':'dtm','datetime_jd':'jd','TDB-UT':'deltat','EclLon':'hc_lon',
                            'EclLat':'hc_lat','r':'hc_rad','ObsEclLon':'gc_lon','ObsEclLat':'gc_lat',
                            'delta':'gc_rad','RA':'ra','DEC':'dec','RA_app':'ra_app','DEC_app':'dec_app',
                            'alpha_true':'phang','illumination':'illum','illum_defect':'illum_def','V':'mag'})
    
    # Convert datetime to "datetime":
    # df.dtm = _pd.to_datetime(df.dtm)  # Only allowed for a limited range!
    df.dtm = df.dtm.str.replace('-Jan-','-01-')
    df.dtm = df.dtm.str.replace('-Feb-','-02-')
    df.dtm = df.dtm.str.replace('-Mar-','-03-')
    df.dtm = df.dtm.str.replace('-Apr-','-04-')
    df.dtm = df.dtm.str.replace('-May-','-05-')
    df.dtm = df.dtm.str.replace('-Jun-','-06-')
    df.dtm = df.dtm.str.replace('-Jul-','-07-')
    df.dtm = df.dtm.str.replace('-Aug-','-08-')
    df.dtm = df.dtm.str.replace('-Sep-','-09-')
    df.dtm = df.dtm.str.replace('-Oct-','-10-')
    df.dtm = df.dtm.str.replace('-Nov-','-11-')
    df.dtm = df.dtm.str.replace('-Dec-','-12-')
    
    return df



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Plot functions for the sluyspy package."""

import matplotlib.pyplot as _plt         # Get matplotlib.pyplot
import numpy.core as _np                 # Get NumPy
from .cli import error as _error


def start_plot(ptype='both', hsize=None,vsize=None, dark_bg=False, xkcd=False, title='Python plot',
               fz=None,lw=None, plot3d=False, man_zorder=True):
    """Start a matplotlib.pyplot plot with a choice of my favourite options.
    
    Parameters:
      ptype (str):        Plot type: 'screen', 'file', 'both' (a compromise) or 'square'.
      hsize (float):      Horizontal size of the figure (hectopixels); can overrule setting by ptype.
      vsize (float):      Vertical size of the figure (hectopixels); can overrule setting by ptype.
    
      dark_bg (bool):     Use a dark background (optional, defaults to False).
      xkcd (bool):        Use XKCD style (optional, defaults to False).
      title (str):        Window title (optional, defaults to 'Python plot').
    
      fz (int):           Font size (optional, defaults to None -> 14 or 16).
      lw (int):           Line width (optional, defaults to None -> 1 or 2).
    
      plot3d (bool):      Start a 3D plot (optional, defaults to False -> 2D).
      map_zorder (bool):  Use manual zorder (defaults to True - opposed to pyplot default: compute).
    
    Returns:
      (tuple):  Tuple containing the plot and axis objects.
    
    Defaults:
      - screen: size 1920x1080, font 14, lw: 1;
      - file:   size 1250x700,  font 14, lw: 2;
      - both:   size 1580x850,  font 16, lw: 2;
      - square: size 850x850,   font 16, lw: 2.
    """
    
    import matplotlib
    
    if xkcd:  _plt.xkcd()                                   # Plot everything that follows in XKCD style
    if dark_bg:  _plt.style.use('dark_background')          # Invert colours
    if xkcd and dark_bg:
        from matplotlib import patheffects
        _plt.rcParams['path.effects'] = [patheffects.withStroke(linewidth=0)]  # Needed for XKCD style on a dark background
    
    if ptype == 'screen':
        if hsize is None: hsize = 19.2  # 1920
        if vsize is None: vsize = 10.8  # 1080
        matplotlib.rcParams.update({'font.size': 14})       # Set font size for all text, aimed at full screen display
    elif ptype == 'file':
        if hsize is None: hsize = 12.5  # 1250
        if vsize is None: vsize =  7.0  # 700
        matplotlib.rcParams.update({'font.size': 14})       # Set font size for all text aimed at an electonic file
        matplotlib.rcParams.update({'lines.linewidth': 2})  # Set default line width to 2
    elif ptype == 'both':
        if hsize is None: hsize = 15.8  # 1580
        if vsize is None: vsize =  8.5  # 850
        matplotlib.rcParams.update({'font.size': 16})       # Set font size for all text: compromise screen visibility and report readability
        matplotlib.rcParams.update({'lines.linewidth': 2})  # Set default line width to 2
    elif ptype == 'square':
        if hsize is None: hsize = 8.5  # 850
        if vsize is None: vsize = 8.5  # 850
        matplotlib.rcParams.update({'font.size': 16})       # Set font size for all text: compromise screen visibility and report readability
    else:
        _error('Unknown plot type: '+ptype+', aborting.')
    
    if fz is not None: matplotlib.rcParams.update({'font.size': fz})        # Set font size for all text
    if lw is not None: matplotlib.rcParams.update({'lines.linewidth': lw})  # Set custom default line width
    fig = _plt.figure(figsize=(hsize,vsize), num=title)  # Custom size
    
    if plot3d:
        ax = fig.add_subplot(111, projection='3d', computed_zorder=(not man_zorder))  # Create a 3D axes object for the current figure
    else:
        ax = fig.add_subplot(111)                                # Create a 2D axes object for the current figure
        
    if xkcd and dark_bg:  ax.spines[:].set_color('white')        # Needed for XKCD style on a dark background
    
    return fig, ax


def hist_norm(obj, x, bins=None, range=None, density=False, weights=None, cumulative=False, bottom=None,
              histtype='bar', align='mid', orientation='vertical', rwidth=None, log=False, color=None, label=None,
              stacked=False, *, data=None, **kwargs):
    """Draw a histogram normalised by total number rather than probability density.
    
    Parameters:
      obj (pyplot object):  Pyplot object or axes object to plot to.
      x (array or sequence of arrays):  Data.
      bins (int or sequence or str):  Number of bins.  Default: rcParams["hist.bins"] (default: 10).
      range (tuple or None):  Bin range, default: None.
      density (bool):  Plot probability density.  Default: False.
      weights (array-like or None):  Weights.  Default: None.
      cumulative (bool or -1):  Plot a cumulative histogram.  Default: False.
      bottom (array-like, scalar, or None):  Location of the bottom of the bins.  Default: None.
      histtype (str):  Histogram type: {'bar', 'barstacked', 'step', 'stepfilled'}, default: 'bar'.
      align (str):  Bin alignment: {'left', 'mid', 'right'}, default: 'mid'.
      orientation (str):  Bin orientation:  {'vertical', 'horizontal'}, default: 'vertical'.
      rwidth (float or None):  The relative width of the bars as a fraction of the bin width, default: None.
      log (bool):  Set the histogram axis to a log scale, default: False.
      color (color or array-like of colors or None):  Color or sequence of colors for bins, one per dataset, default: None.
      label (str or None):  Legend label, default: None.
      stacked (bool):  Stack multiple data on top of each other, rather than side by side, default: False.
      data (indexable object, optional):  If given, the following parameters also accept a string s, which is interpreted as data[s] (unless this raises an exception): x, weights.
    
    
    Returns:
      (tuple):  Tuple consisting of (n, bin_edges, patches):
      
      - (array or list of arrays):  The values of the histogram bins.
      - (array): The edges of the bins. Length nbins + 1.
      - (BarContainer or list of Polygons):  (List of) container(s) of individual artists used to create the histogram(s).
    """
        
    weights = _np.ones_like(x) / len(x)  # Set all weights to 1/n to pass them to pyplot.hist()
    
    n, bin_edges, patches = obj.hist(x, bins=bins, range=range, density=density, weights=weights,
                                     cumulative=cumulative, bottom=bottom, histtype=histtype, align=align,
                                     orientation=orientation, rwidth=rwidth, log=log, color=color,
                                     label=label, stacked=stacked, data=data, **kwargs)
    
    return n, bin_edges, patches


def show_ctrlc():
    """Call pyplot.show() and catch Ctrl-C to exit gently."""
    
    try:
        _plt.show()
    except KeyboardInterrupt:
        print(' - Received keyboard interrupt whilst showing plot, aborting.')  # " - " to create some space after "^C"
        exit(0)
        
    return


def pause_ctrlc(interval):
    """Call pyplot.pause() and catch Ctrl-C to exit gently.
    
    Parameters:
      interval (float):  pause time in seconds.
    """
    
    try:
        _plt.pause(interval)
    except KeyboardInterrupt:
        print(' - Received keyboard interrupt whilst pausing plot, aborting.')  # " - " to create some space after "^C"
        exit(0)
        
    return


def arrow_head_between_points(ax, xx,yy, dx=0,dy=0, scale=1, **kwargs):
    """Plot an arrow head between two 2D points.

    Parameters:
      ax (axes):      Axes object to plot on.
      xx (float):     Array with two x values, for arrow-head position and direction.
      yy (float):     Array with two y values, for arrow-head position and direction.
      dx (float):     Shift arrow head in x direction.
      dy (float):     Shift arrow head in y direction.
      scale (float):  Scale arrow head.
    """
    
    # Local scaling parameter: undo arrow-head scaling with length:
    lscale = 0.1 * scale / _np.sqrt((xx[1]-xx[0])**2 + (yy[1]-yy[0])**2)
    
    if len(xx)!=2: _error('arrow_head_between_points(): The input variable xx must have two elements')
    if len(yy)!=2: _error('arrow_head_between_points(): The input variable yy must have two elements')
    
    ax.quiver(xx[1]+dx, yy[1]+dy, xx[1]-xx[0], yy[1]-yy[0],
              angles='xy', scale_units='width', pivot='mid',
              scale=3/lscale, width=5e-3*lscale,
              headwidth=10*lscale, headlength=15*lscale, headaxislength=10*lscale,
              **kwargs)
    return



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Weather functions for the sluyspy package"""


import numpy as _np
import pandas as _pd
import datetime as _dt
import astroconst as _ac



def wind_chill_temperature(temp, wind_vel):
    """Compute wind-chill ("real-feel") temperature from the air temperature and wind velocity.
    
    Parameters:
      temp (float):      Air temperature at 1.5m above ground (°C).
      wind_vel (float):  Wind velocity at 10m above ground (m/s).
    
    Returns:
      (float):  Wind-chill temperature (°C).
    
    Source:
      - https://en.wikipedia.org/wiki/Wind_chill
        - but wind speed in m/s instead of km/h (3.6^0.16*11.37=13.96; 3.6^0.16*39.65=48.67).
        - ensure that the wind chill NEVER exceeds the temperature (was: only if wind_vel < 1.3).
      - note: the power 0.16 converts 10m wind velocity to 1.5m.
    """
    
    # Compute wind-chill ("real-feel") temperature:
    wchil = 13.12 + 0.6215 * temp + (0.4867*temp - 13.96)*wind_vel**0.16
    
    # Wind chill == air temperature for T>10°C:
    if _np.ndim(wchil) == 0:  # Dimension: 0: scalar, >0: array-like
        # if wind_vel < 1.3:
        wchil = min(wchil, temp)
    else:
        # wchil[wind_vel < 1.3] = _np.minimum( wchil[wind_vel < 1.3], temp[wind_vel < 1.3] )
        wchil = _np.minimum( wchil, temp )
    
    # Round off to nearest 0.1°C, since it is unlikely that the air temperature is known to better than 0.1°C
    # or that this simple model is more accurate than that:
    wchil = _np.round(wchil,1)
    
    return wchil



def wind_dir_str_from_az(az):
    """Convert a wind direction (azimuth) in degrees to a three-character English abbreviation string.
    
    Parameters:
      az (float):  Azimuth (degrees; 0=North, 90=East).
    
    Returns:
      (str):  Three-character English wind-direction abbreviation.
    """
    
    tmp_val = int((az/22.5)+0.5)
    str_arr = ['N','NNE','NE','ENE','E','ESE', 'SE', 'SSE','S','SSW','SW','WSW','W','WNW','NW','NNW']
    
    return str_arr[(tmp_val % 16)]



def sky_power_from_rain_means(rain):
    print('Sorry, this function moved to sluyspy.solar_energy._cloud_power_from_rain(), called from solar_power_from_true_sky_rain()')
    exit(1)
    return


def sky_power_from_rain_medians(rain):
    print('Sorry, this function moved to sluyspy.solar_energy._cloud_power_from_rain(), called from solar_power_from_true_sky_rain()')
    exit(1)
    return


def dew_point_from_tempc_rh(temp_c, rh):
    """Compute the dew point from the temperature and relative humidity.
    
    Parameters:
      temp_c (float):  Air temperature (degrees Celsius).
      rh (float):      Relative humidity (fraction).
    
    Returns:
      (float):  Dew point (degrees Celsius).
    
    See:
      https://en.wikipedia.org/wiki/Dew_point
    """
    
    temp_c0  = 273.7
    tempfac  = 17.27
    gam = tempfac * temp_c/(temp_c0+temp_c) + _np.log(rh)
    
    return temp_c0 * gam/(tempfac-gam)


def water_vapor_saturated_density_from_tempc(temp_c):
    """Compute the saturated water-vapor density in air for a given temperature.
    
    Parameters:
      temp_c (float):  Air temperature (degrees Celsius).
    
    Returns:
      (float):  Saturated water-vapor density (g/m^3).
    
    Note:
      - uses data from http://hyperphysics.phy-astr.gsu.edu/hbase/kinetic/relhum.html#c3:
        - for T= -10 - +60°C.
      - drops to 0 below T~-25°C;
      - significantly better results than original fit (3rd order polynomial):
        - for T<~-15°C and T>~45°C: by eye;
        - for -10°C < T < 60°C (range original fit -0°C < T < 40°C):
          - typical sigma 0.116 (T=0-40°C) -> 0.030 g/m^3 (T=-10-60°C);
          - max. abs. deviation: 0.168 g/m^3 (T=0-40°C; @T=0°C)  ->  0.0445 g/m^3 (T=-10-60°C; @T=25°C)
          - max. rel. deviation: 3.46  %     (T=0-40°C; @T=0°C)  ->  0.194  %     (T=-10-60°C; @T=25°C)
    """
    
    return _np.maximum(4.85684 + 3.32664e-01 * temp_c + 1.00885e-02 * temp_c**2 + 1.89345e-04 * temp_c**3
                       + 1.09606e-06 * temp_c**4 + 1.83396e-08 * temp_c**5, 0)


def absolute_humidity_from_tempc_rh(temp_c, rh):
    """Compute the absolute humidity (water-vapor density) from the temperature and relative humidity.
    
    Parameters:
      temp_c (float):  Air temperature (degrees Celsius).
      rh (float):      Relative humidity (fraction).
    
    Returns:
      (float):  Absolute humidity (water-vapor density;  g/m^3).
    """
    
    return water_vapor_saturated_density_from_tempc(temp_c) * rh


def knmi_read_hourly_weather(filename, start_date=None, end_date=None, tidy=True):
    """Reads a KNMI file containing hourly weather measurement data from a particular station.
    
    Reads the CSV files provided on the KNMI website for the weather data over the last decade from a chosen
    station. This function reads the weather data from a chosen start date to the chosen end date, which is
    NOT included.  For example, for the whole of the year 2020, specify 2020,1,1, 2021,1,1.
    
    Parameters:
      filename (str): filename to select the specific station, this is "uurgeg_260_2011-2020_Bilt.csv" for De Bilt
    
      start_date (datetime):  start date:  date of the first data entry to be returned (timezone aware; optional).
      end_date   (datetime):  end date:    date of the first data entry NOT to be returned (timezone aware; optional).
    
      tidy       (bool):      tidy up: remove station column, convert silly date and time to datetime,
                              temperature to °C and sunshine to a fraction (optional).
    
    Returns:
      pandas.DataFrame:  Table containing ~23-25 columns with weather data.
    
      - yyyymmdd (int):  Date (YYYY=year,MM=month,DD=day).
      - hh (int):        Time in hours (note: 1-24!), the hourly division 05 runs from 04.00 UT to 5.00 UT.
      - dtm (datetime):  Date and time of the entry (UT), replacing yyyymmdd and hh.
    
      - t (float):       Temperature at 1.50 m at the time of observation (°C).
      - sq (float):      Fraction of the hour with sunshine (0-1).
      - q (float):       Global horizontal radiation (W/m^2 (or J/cm^2/h)).
    """
    
    # Read the data file into a Pandas DataFrame.  Use lower-case column names:
    col_names = ['stn','yyyymmdd','hh','dd','fh','ff','fx','t','t10n','td','sq','q','dr','rh','p','vv','n','u','ww','ix','m','r','s','o','y']
    try:
        knmi_data = _pd.read_csv(filename, header=28, names=col_names, sep=r'\s*,\s*', engine='python')
    except Exception as e:
        print(e)
        exit(1)
    
    # Select the data for the desired date range.  Add one hour, since datetime contains the time at which the hour ENDS!
    if start_date and end_date:
        knmi_data = knmi_data[knmi_data['datetime'] >= start_date + _dt.timedelta(hours=1)]  # Date must be >= start_date
        knmi_data = knmi_data[knmi_data['datetime'] <    end_date + _dt.timedelta(hours=1)]  # Date must be < end_date
        knmi_data = knmi_data.reset_index(drop=True)  # Reset the index so that it starts at 0 again.  Don't keep the original index as a column.
    
    # Tidy up if desired:
    if tidy:
        del knmi_data['stn']
        # Convert the yyyymmdd and hh columns into a datetime-object column with hours in [0,23] rather than in [1,24]:
        knmi_data = knmi_datetime_from_datehour(knmi_data)
        
        knmi_data.loc[:, 't']  /= 10                              # Convert the temperature at 1.50 m from 0.1°C -> °C.
        knmi_data.loc[knmi_data.loc[:, 'sq'] == -1, 'sq'] = 0.25  # Sunshine time == -1 indicates 0 - 0.05 hours, so assume 0.025 hours, expressed in [0.1 hours]
        knmi_data.loc[:, 'sq'] /= 10                              # Convert the sunshine time from [0.1 hours] to a fraction (0-1)
        knmi_data.loc[:, 'q']  /= 0.36                            # Convert the global horizontal radiation from [J/cm^2/h] to [W/m^2]
    
    return knmi_data


def knmi_datetime_from_datehour(knmi_data):
    """Return a datetime column named 'dtm' for given KNMI date and hour columns.
    
    The KNMI date is expressed as an integer formatted as YYYYMMDD, while the hours run from 1-24 rather than
    from 0-23.  This causes problems when converting to Python or Pandas datetime objects.
    
    Parameters:
      knmi_data (Pandas df):  KNMI weather dataframe.
    
    Returns:
      (Pandas df):  KNMI weather dataframe.
    """
    
    from astrotool.date_time import fix_date_time
    
    # Split the yyyymmdd column into separate numpy arrays:
    ymd     = knmi_data['yyyymmdd'].values  # Numpy array
    years   = _np.floor(ymd/1e4).astype(int)
    months  = _np.floor((ymd - years*1e4)/100).astype(int)
    days    = _np.floor(ymd - years*1e4 - months*100).astype(int)
    
    # Create numpy arrays for the time variables:
    hours   = knmi_data['hh'].values  # Numpy array
    minutes = _np.zeros(hours.size)
    seconds = _np.zeros(hours.size) + 0.001  # 1 ms past the hour, to ensure no negative round-off values occur (e.g. 2021,1,1, 0,0,-1e-5 -> 2020,12,31, 23,59,59.99999)
    
    # Fix the dates, e.g. turning 2020-12-31 24:00:00 to 2021-01-01 00:00:00:
    years,months,days, hours,minutes,seconds = fix_date_time(years,months,days, hours,minutes,seconds)
    
    # Combine the 1D numpy arrays into a single 2D array with the original arrays as COLUMNS, and convert it to a Pandas df:
    dts = _pd.DataFrame(_np.vstack([years,months,days,hours]).transpose(), columns=['year','month','day','hour'])
    dts = _pd.to_datetime(dts, utc=True)    # Turn the columns in the df into a single datetime64[ns] column
    
    # Add the datetime column to the KNMI weather dataframe:
    knmi_data['dtm'] = dts
    knmi_data = knmi_data[['dtm'] + [x for x in knmi_data.columns if x != 'dtm']]  # Move dtm column to front
    
    del knmi_data['yyyymmdd'], knmi_data['hh']
    
    return knmi_data


def knmi_station_data(remove_nans=False):
    """Return a Pandas DataFrame containing the data of Dutch KNMI weather stations.
    
    Parameters:
      remove_nans (bool):  Remove entries containing a NaN.  Optional; defaults to False.
    
    Returns:
      (pd.df):  Pandas DataFrame containing and index and five columns:
    
      - index (int):  Station number. Note that most stations start with a leading zero, which is omitted here.
      - lon (float):  Geographical longitude of the station, in radians east of Greenwich.
      - lat (float):  Geographical latitude of the station, in radians north of the equator.
      - alt (float):  Altitude of the station above sea level, in metres.
      - name (str):   (A) name of the station.
      - fname (str):  String to compose a file name from.  This is the name from the KNMI .nc files,
                      in lower case, with spaces and slashes replaced by underscores, and brackets removed.
    """
    
    df = _pd.DataFrame( {
        'lon':   {6201: float('nan'), 6203: float('nan'), 6204: float('nan'), 6205: float('nan'), 6207: float('nan'), 6208: float('nan'), 6209: 4.518, 6210: 4.43, 6211: float('nan'), 6214: float('nan'), 6215: 4.437, 6216: float('nan'), 6225: 4.555, 6229: float('nan'), 6233: float('nan'), 6235: 4.781, 6236: float('nan'), 6237: float('nan'), 6238: float('nan'), 6239: float('nan'), 6240: 4.79, 6242: 4.921, 6248: 5.174, 6249: 4.979, 6251: 5.346, 6252: float('nan'), 6257: 4.603, 6258: 5.401, 6260: 5.18, 6265: 5.274, 6267: 5.384, 6269: 5.52, 6270: 5.752, 6273: 5.888, 6275: 5.873, 6277: 6.2, 6278: 6.259, 6279: 6.574, 6280: 6.585, 6283: 6.657, 6285: 6.399, 6286: 7.15, 6290: 6.891, 6308: 3.379, 6310: 3.596, 6311: 3.672, 6312: 3.622, 6313: 3.242, 6315: 3.998, 6316: 3.694, 6317: float('nan'), 6319: 3.861, 6320: float('nan'), 6321: float('nan'), 6323: 3.884, 6324: 4.006, 6330: 4.122, 6331: 4.193, 6340: 4.342, 6343: 4.313, 6344: 4.447, 6348: 4.926, 6350: 4.936, 6356: 5.146, 6370: 5.377, 6375: 5.707, 6377: 5.763, 6380: 5.762, 6391: 6.197},
        'lat':   {6201: float('nan'), 6203: float('nan'), 6204: float('nan'), 6205: float('nan'), 6207: float('nan'), 6208: float('nan'), 6209: 52.465, 6210: 52.171, 6211: float('nan'), 6214: float('nan'), 6215: 52.141, 6216: float('nan'), 6225: 52.463, 6229: float('nan'), 6233: float('nan'), 6235: 52.928, 6236: float('nan'), 6237: float('nan'), 6238: float('nan'), 6239: float('nan'), 6240: 52.318, 6242: 53.241, 6248: 52.634, 6249: 52.644, 6251: 53.392, 6252: float('nan'), 6257: 52.506, 6258: 52.649, 6260: 52.1, 6265: 52.13, 6267: 52.898, 6269: 52.458, 6270: 53.224, 6273: 52.703, 6275: 52.056, 6277: 53.413, 6278: 52.435, 6279: 52.75, 6280: 53.125, 6283: 52.069, 6285: 53.575, 6286: 53.196, 6290: 52.274, 6308: 51.381, 6310: 51.442, 6311: 51.379, 6312: 51.768, 6313: 51.505, 6315: 51.447, 6316: 51.657, 6317: float('nan'), 6319: 51.226, 6320: float('nan'), 6321: float('nan'), 6323: 51.527, 6324: 51.596, 6330: 51.992, 6331: 51.48, 6340: 51.449, 6343: 51.893, 6344: 51.962, 6348: 51.97, 6350: 51.566, 6356: 51.859, 6370: 51.451, 6375: 51.659, 6377: 51.198, 6380: 50.906, 6391: 51.498},
        'alt':   {6201: float('nan'), 6203: float('nan'), 6204: float('nan'), 6205: float('nan'), 6207: float('nan'), 6208: float('nan'), 6209: 0.0, 6210: -0.2, 6211: float('nan'), 6214: float('nan'), 6215: -1.1, 6216: float('nan'), 6225: 4.4, 6229: float('nan'), 6233: float('nan'), 6235: 1.2, 6236: float('nan'), 6237: float('nan'), 6238: float('nan'), 6239: float('nan'), 6240: -3.3, 6242: 10.8, 6248: 0.8, 6249: -2.4, 6251: 0.7, 6252: float('nan'), 6257: 8.5, 6258: 7.3, 6260: 1.9, 6265: 13.9, 6267: -1.3, 6269: -3.7, 6270: 1.2, 6273: -3.3, 6275: 48.2, 6277: 2.9, 6278: 3.6, 6279: 15.8, 6280: 5.2, 6283: 29.1, 6285: 0.0, 6286: -0.2, 6290: 34.8, 6308: 0.0, 6310: 8.0, 6311: 0.0, 6312: 0.0, 6313: 0.0, 6315: 0.0, 6316: 0.0, 6317: float('nan'), 6319: 1.7, 6320: float('nan'), 6321: float('nan'), 6323: 1.4, 6324: 0.0, 6330: 11.9, 6331: 0.0, 6340: 19.2, 6343: 3.5, 6344: -4.3, 6348: -0.7, 6350: 14.9, 6356: 0.7, 6370: 22.6, 6375: 22.0, 6377: 30.0, 6380: 114.3, 6391: 19.5},
        'name':  {6201: float('nan'), 6203: float('nan'), 6204: float('nan'), 6205: float('nan'), 6207: float('nan'), 6208: float('nan'), 6209: 'IJmond', 6210: 'Valkenburg Zh', 6211: float('nan'), 6214: float('nan'), 6215: 'Voorschoten', 6216: float('nan'), 6225: 'IJmuiden', 6229: float('nan'), 6233: float('nan'), 6235: 'De Kooy', 6236: float('nan'), 6237: float('nan'), 6238: float('nan'), 6239: float('nan'), 6240: 'Schiphol', 6242: 'Vlieland', 6248: 'Wijdenes', 6249: 'Berkhout', 6251: 'Hoorn Terschelling', 6252: float('nan'), 6257: 'Wijk aan Zee', 6258: 'Houtribdijk', 6260: 'De Bilt', 6265: 'Soesterberg', 6267: 'Stavoren', 6269: 'Lelystad', 6270: 'Leeuwarden', 6273: 'Marknesse', 6275: 'Deelen', 6277: 'Lauwersoog', 6278: 'Heino', 6279: 'Hoogeveen', 6280: 'Eelde', 6283: 'Hupsel', 6285: 'Huibertgat', 6286: 'Nieuw Beerta', 6290: 'Twenthe', 6308: 'Cadzand', 6310: 'Vlissingen', 6311: 'Hoofdplaat', 6312: 'Oosterschelde', 6313: 'Vlakte van De Raan', 6315: 'Hansweert', 6316: 'Schaar', 6317: float('nan'), 6319: 'Westdorpe', 6320: float('nan'), 6321: float('nan'), 6323: 'Wilhelminadorp', 6324: 'Stavenisse', 6330: 'Hoek van Holland', 6331: 'Tholen', 6340: 'Woensdrecht', 6343: 'Rotterdam Geulhaven', 6344: 'Rotterdam', 6348: 'Cabauw Mast', 6350: 'Gilze-Rijen', 6356: 'Herwijnen', 6370: 'Eindhoven', 6375: 'Volkel', 6377: 'Ell', 6380: 'Maastricht', 6391: 'Arcen'},
        'fname': {6201: 'd15-fa-1', 6203: 'p11-b', 6204: 'k14-fa-1c', 6205: 'a12-cpp', 6207: 'l9-ff-1', 6208: 'awg-1', 6209: 'ijmond_wp', 6210: float('nan'), 6211: 'j6-a', 6214: 'bg-ohvs2', 6215: 'voorschoten_aws', 6216: 'hollandse_kust_zuid_alfa_hkza', 6225: 'ijmuiden_wp', 6229: 'texelhors_wp', 6233: 'assendelft_btp', 6235: 'de_kooy_vk', 6236: 'muiden_btp', 6237: 'nieuw_vennep_btp', 6238: 'nieuwkoop_btp', 6239: 'f3-fb-1', 6240: 'amsterdam_schiphol_ap', 6242: 'vlieland', 6248: 'wijdenes_wp', 6249: 'berkhout_aws', 6251: 'terschelling_hoorn_aws', 6252: 'k13-a', 6257: 'wijk_aan_zee_aws', 6258: 'houtribdijk_wp', 6260: 'de_bilt_aws', 6265: float('nan'), 6267: 'stavoren_aws', 6269: 'lelystad_ap', 6270: 'leeuwarden', 6273: 'marknesse_aws', 6275: 'deelen', 6277: 'lauwersoog_aws', 6278: 'heino_aws', 6279: 'hoogeveen_aws', 6280: 'groningen_ap_eelde', 6283: 'hupsel_aws', 6285: 'huibertgat_wp', 6286: 'nieuw_beerta_aws', 6290: 'twenthe_aws', 6308: 'cadzand_wp', 6310: 'vlissingen_aws', 6311: float('nan'), 6312: 'oosterschelde_wp', 6313: 'vlakte_van_de_raan', 6315: 'hansweert', 6316: 'oosterschelde_4', 6317: 'borssele_alfa_bsa', 6319: 'westdorpe_aws', 6320: 'lichteiland_goeree', 6321: 'euro_platform', 6323: 'wilhelminadorp_aws', 6324: 'stavenisse', 6330: 'hoek_van_holland_aws', 6331: 'marollegat', 6340: 'woensdrecht', 6343: 'rotterdam_geulhaven', 6344: 'rotterdam_the_hague_ap', 6348: 'cabauw_tower_aws', 6350: 'gilze_rijen', 6356: 'herwijnen_aws', 6370: 'eindhoven_ap', 6375: 'volkel', 6377: 'ell_aws', 6380: 'maastricht_aachen_ap', 6391: 'arcen_aws'}
    } )
    
    # Convert coordinates from degrees to radians:
    df.lon *= _ac.d2r
    df.lat *= _ac.d2r
    
    if remove_nans:
        df = df[df['lon'].notna()]      # Keep only rows where df.col is NOT Null/None/NaN
        df = df[df['fname'].notna()]    # Keep only rows where df.col is NOT Null/None/NaN
    
    return df



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'sluyspy'
copyright = '2023, Marc van der Sluys'
author = 'Marc van der Sluys'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.imgmath',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
]

# templates_path = ['_templates']
exclude_patterns = ['build', 'Thumbs.db', '.DS_Store']

language = 'en'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
# html_static_path = ['_static']

# -- Options for intersphinx extension ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}

# -- Options for todo extension ----------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/todo.html#configuration

todo_include_todos = True



# MvdS: extra options for Napoleon
napoleon_google_docstring = True
napoleon_use_param = False



# The master toctree document:
master_doc = 'index'



# -- MvdS: Options for LaTeX output ------------------------------------------------
latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    'pointsize': '11pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    'figure_align': 'htbp',
}



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Solar-energy functions for the sluyspy package"""


import numpy as _np
import pandas as _pd
# import pytz as _tz
# import solarenergy as _se


def solar_power_from_true_sky_rain(Pclear, cloud_cover, rain, fittype='means'):
    """Compute solar power coming from a realistic (non-clear) sky, as a function of the power from the clear sky, the cloud cover and rain.
    
    Parameters:
      Pclear (float):       (Electrical) solar power from clear sky (e.g. W or kW).
      cloud_cover (float):  Cloud cover (%).
      rain (float):         Rain (mm/h).
      fittype (str):        Type of fit: 'orig_data' (worst), 'means' (best, default) or 'medians'.
    
    Returns:
      (float):  (Electrical) solar power coming from a realistic sky (clear + clouded; same unit as Pclear).
    """
    
    # Set up df:
    df = _pd.DataFrame()
    df['rain_int'] = rain
    df['clouds']   = cloud_cover
    df['Pclrsky']  = Pclear
    
    # Rain intensity *if* it rains := rain this hour / cloud cover (assumption: no clouds, no rain):
    df.loc[df.clouds>0, 'rain_int'] = df.loc[df.clouds>0, 'rain_int'] / df.loc[df.clouds>0, 'clouds'] * 100
    
    df['PPclr'] = _cloud_power_from_rain(df.rain_int, fittype)  # P/P_clearsky, based on rain
    
    df.PPclr = (df.PPclr * df.clouds  +  1 * (100-df.clouds))/100  # Weighted sum of cloud and sun contributions
    df['Pclouds'] = df.Pclrsky*df.PPclr  # Take into account clouds in predicted power
    
    # return df.rain_int, df.PPclr, df.Pclouds
    return df.Pclouds

    
def _cloud_power_from_rain(rain, fittype='means'):
    """Guestimate the fraction of light power coming from an overcast sky, from the amount of rain.
    
    Parameters:
      rain (float):   Rain value(s), in mm/hr.
      fittype (str):  Type of fit: 'means' (best?; default) or 'medians'.
    
    Returns:
      (float):  P/Pclear; the fraction of clear-sky solar power coming from an overcast sky.
    
    Note:
      - this is about as close to witchcraft as it is to science;
      - this version uses a fit to the MEANS of the rain bins;
      - see compare-selections.py.
    """
    
    if fittype == 'means':
        # 2023-08: red.chi2 ~ 0.191161
        # ppclr = 20 * _np.square(rain) - 4 * rain + 0.395  # 0-0.1
        # ppclr[rain>0.1] = 0.74 * _np.exp( -( rain[rain>0.1] + 2.97 ) / 2.3 )
        
        # 2023-08: red.chi2 ~ 0.190071
        ppclr = 80 * _np.square(rain) - 8 * rain + 0.4 - 0.0009  # 0-0.05
        ppclr[rain>0.05] = 0.74 * _np.exp( -( rain[rain>0.05] + 2.97 ) / 2.3 )
        
    elif fittype == 'medians':
        # 2023-08: red.chi2 ~ 0.291121
        # ppclr = 13 * _np.square(rain) - 2.6 * rain + 0.25  # 0-0.1
        # ppclr[rain>0.1] = 2.06 * _np.exp( -( rain[rain>0.1] + 10.67 ) / 3.78 )
        
        # 2023-08: red.chi2 ~ 0.292222
        ppclr = 52 * _np.square(rain) - 5.2 * rain + 0.25 + 0.00085  # 0-0.05
        ppclr[rain>0.05] = 2.06 * _np.exp( -( rain[rain>0.05] + 10.67 ) / 3.78 )
    
    return ppclr


def solar_power_from_true_sky_rh(Pclrsky, cloud_cover, rh, fittype='means'):
    """Compute solar power coming from a realistic (non-clear) sky, as a function of the power from
       the clear sky, the cloud cover and relative humidity.
    
    Parameters:
      Pclrsky (float):      (Electrical) solar power from clear sky (e.g. W or kW).
      cloud_cover (float):  Cloud cover (%).
      rh (float):           Relative humidity (%).
      fittype (str):        Type of fit: 'orig_data' (worst), 'means' (best, default) or 'medians'.
    
    Returns:
      (float):  (Electrical) solar power coming from a realistic sky (clear + clouded; same unit as Pclear).
    """
    
    clouds = cloud_cover/100  # % -> fraction
    
    PPclr_cloud, PPclr_sun = _cloud_power_from_rh_means(rh, fittype)  # P/P_clearsky, based on rh
    
    PPclr = (PPclr_cloud * clouds  +  PPclr_sun * (1-clouds))  # Weighted sum of cloud and sun contributions
    Pclouds = Pclrsky*PPclr  # Take into account clouds in predicted power
    
    return Pclouds


def _cloud_power_from_rh_means(rh, fittype='means'):
    """Guestimate the fraction of light power coming from an overcast sky, from the relative humidity.
    
    Parameters:
      rh (float):     Relative humidity (%).
      fittype (str):  Type of fit: 'orig_data' (worst), 'means' (best) or 'medians'.
    
    Returns:
      (tuple):  Tuple containing (ppclr_cloud, ppclr_sun):
    
      - ppclr_cloud (float):  P/Pclear; the fraction of clear-sky solar power coming from an overcast sky.
      - ppclr_sun (float):    P/Pclear; the fraction of clear-sky solar power coming from a sunny sky.
    
    Note:
      - this is about as close to witchcraft as it is to science;
      - see compare-selections.py.
    """
    
    if fittype == 'orig_data':
        ppclr_cloud = 0.525804 - 5.09361e-03*rh  # When clouded
        ppclr_sun   = 0.977976 + 5.09722e-04*rh  # When sunny
    if fittype == 'means':  # Looks best
        ppclr_cloud = 0.973944 - 8.74388e-03*rh  # When clouded
        ppclr_sun   = 0.908390 + 2.13867e-03*rh  # When sunny
    if fittype == 'medians':
        ppclr_cloud = 0.885572 - 8.40989e-03*rh  # When clouded
        ppclr_sun   = 0.950993 + 1.31021e-03*rh  # When sunny
    
    return ppclr_cloud, ppclr_sun



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Command-line interface functions for the sluyspy package"""

import sys as _sys
from termcolor import colored as _clr


def dialog(text):
    """Present a dialog text and wait for a single-key answer.
    
    Parameters:
      text (str):  The text to print.
    
    Returns:
      (str):  The single character typed by the user.
    """
    
    import getch
    
    _sys.stdout.write(text+' ')  # No newline
    _sys.stdout.flush()          # Show the previous line
    
    char = getch.getche()  # Ask for user input, displayed on the screen
    print()                # Newline after input
    
    return char


def error(message, exit_program=True, exit_code=1):
    """Print a coloured error message to stderr and exit.
    
    Parameters:
      message (str):        Message to print.
      exit_program (bool):  Exit the program or not, defaults to True.
      exit_code (int):      Exit code to exit the program with, defaults to 1.
    """
    
    _sys.stderr.write('\n'+_clr('ERROR: '+str(message), 'white', 'on_red', attrs=['bold'])+'\n\n')
    
    if exit_program: exit(exit_code)
    return


def warn(message, exit_program=False, exit_code=1):
    """Print a coloured warn message to stderr and exit.
    
    Parameters:
      message (str):        Message to print.
      exit_program (bool):  Exit the program or not, defaults to False.
      exit_code (int):      Exit code to exit the program with, defaults to 1.
    """
    
    _sys.stderr.write('\n'+_clr('Warning: '+str(message), 'white', 'on_yellow', attrs=['bold'])+'\n\n')
    
    if exit_program: exit(exit_code)
    return



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Numerics stuff for the sluyspy package"""


import numpy as _np
import astroconst as _ac


eps  = _np.finfo(_np.float64).eps;  """Smallest value for which 1+eps != 1:  2.2204460492503131e-16"""
eps1 = 1 + eps;                     """Smallest value larger than 1:  1 + 2.2204460492503131e-16"""
tiny = _np.nextafter(0, 1);         """Smallest value larger than 0: 4.9406564584124654e-324"""


def sigdig(num, dig=14):
    """Return a string with a given number using (at most) the specified number of significant digits.
    
    Parameters:
      num (float):  Number to print.
      dig (int):    Number of significant digits to use (optional; defaults to 14 to avoid machine rounding).
    
    Returns:
      (struct):  Struct containing command-line arguments.
    
    Note:
      - Trailing zeros are NOT printed - hence, this function does not EXACTLY print the desired number of
        significant digits;
      - 0.075 with a single digit is printed as 0.07, hence multiply with eps1!
    
    See:
      E.g. the discussions here: https://stackoverflow.com/q/3410976/1386750
    """
    
    if num is None:
        sigdig = 'None'
    else:
        fmt = '%%0.%ig' % (dig)
        sigdig = fmt % (num * eps1)  # w/o snum.eps1, (0.075,1) will be written as 0.07!
        
    return sigdig


def map_pi_pi(ang):
    """Map an (array of) angle(s) to lie between -PI and +PI.

    Parameters:
      ang (float):  (Array of) angle(s).
    
    Returns:
      (float):  (Array) of angles between -PI and +PI.
    """
    
    return _np.mod(ang+_ac.pi, _ac.pi2)-_ac.pi


def nint(val):
    """Convert a float to the nearest integer with proper rounding.
    
    Parameters:
      val (float):  Value to convert.
    
    Returns:
      (int):  Nearest integer (or float('nan')).
    """
    
    if _np.isnan(val):
        nint = float('nan')
    else:
        nint = int(_np.round(val,0))
        
    return nint



#!/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Environment functions for the sluyspy package."""


from dataclasses import dataclass as _dataclass

from sluyspy import system as _ssys
import astroconst as _ac


@_dataclass
class Environment:
    tz:       str  = '';     """Time zone"""
    geo_lon:  float = 0.0;   """Geographical longitude in radians east of Greenwich"""
    geo_lat:  float = 0.0;   """Geographical latitude in radians north of the equator"""
    geo_alt:  float = 0.0;   """Geographical altitude in metres above sea level"""
    
    host:     str  = '';     """Host name"""
    home:     str  = '';     """Home directory"""
    on_zotac: bool = False;  """Am I on Zotac?"""
    on_think: bool = False;  """Am I on Think?"""
    
    sp_dir:          str = '';  """Solar-panel directory"""
    el_dir:          str = '';  """Electricity-meter directory"""
    
    knmi_10min_dir:  str = '';  """Directory for 10-min KNMI data"""
    knmi_hourly_dir: str = '';  """Directory for hourly KNMI data"""
    knmi_daily_dir:  str = '';  """Directory for daily KNMI data"""
    wpw_dir:         str = '';  """WP weather directory"""
    
    hwc_dir:         str = '';  """HWC main directory"""
    
    
def environment(cfg_file='.python_environment.cfg'):
    """Return my computing environment.
    
    Parameters:
      cfg_file (str):  Configuration file to read system environment from (relative to home directory).
    
    Returns:
      (Environment):  Dataclass containing the environment settings.
    """
    
    env = Environment()
    env.host = _ssys.host()
    env.home = _ssys.homedir()
    
    env.on_zotac = env.host == 'zotac'
    env.on_think = env.host == 'think'
    
    
    # Read system config file:
    import configparser
    config = configparser.ConfigParser(inline_comment_prefixes=('#'))
    config.read(env.home+'/'+cfg_file)
    
    # Section Localisation:
    env.tz      = config.get('Localisation', 'timezone',  fallback=env.tz)       # My timezone
    env.geo_lon = config.getfloat('Localisation', 'longitude', fallback=env.geo_lon)  # My longitude
    env.geo_lat = config.getfloat('Localisation', 'latitude',  fallback=env.geo_lat)  # My latitude
    env.geo_alt = config.getfloat('Localisation', 'altitude',  fallback=env.geo_alt)  # My altitude
    
    env.geo_lon *= _ac.d2r  # Convert from degrees to radians
    env.geo_lat *= _ac.d2r
    
    # Section SolarPanels:
    env.sp_dir = config.get('SolarPanels', 'basedir', fallback=env.sp_dir).replace('~', env.home)  # SP base dir
    
    # Section ElectricityMeter:
    env.el_dir = config.get('ElectricityMeter', 'basedir', fallback=env.el_dir).replace('~', env.home)  # EM base dir
    
    # Section Weather:
    env.knmi_10min_dir  = config.get('Weather', 'knmi_10min_dir',  fallback=env.knmi_10min_dir).replace('~',  env.home)  # KNMI 10-min dir
    env.knmi_hourly_dir = config.get('Weather', 'knmi_hourly_dir', fallback=env.knmi_hourly_dir).replace('~', env.home)  # KNMI hourly dir
    env.knmi_daily_dir  = config.get('Weather', 'knmi_daily_dir',  fallback=env.knmi_daily_dir).replace('~', env.home)   # KNMI daily dir
    env.wpw_dir         = config.get('Weather', 'wpw_dir',         fallback=env.wpw_dir).replace('~', env.home)          # WP weather dir
    
    # Section HWC:
    env.hwc_dir         = config.get('HWC', 'hwc_dir',             fallback=env.hwc_dir).replace('~', env.home)              # HWC main dir
    
    return env


if __name__ == '__main__':
    print(environment())



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""System functions for the sluyspy package."""


from pathlib import Path as _Path
import socket as _socket


def host():
    """Return the current host name.
    
    Returns:  
      (str):  The current host name.
    """
    
    return _socket.gethostname()


def homedir():
    """Return my home directory as a string without trailing slash.
    
    Returns:  
      (str):  My home directory as a string without trailing slash.
    """
    
    return str(_Path.home())


def on_zotac():
    """Tell whether I am on a host called zotac.
    
    Returns:
      (bool):  True when I am on a host called zotac, False otherwise.
    """
    
    return host() == 'zotac'


def temp_file_name(dir_name=None, base_name='.tmpfile', ext='tmp'):
    """Return a temporary file name, using the system clock, like "dir/file_YYYYMMDD-HHMMSS.mmmmmm.ext".
    
    Parameters:
      dir_name (str):   Name of the directory for the file to be in.  Use homedir if None.
      base_name (str):  Base name of the file.
      ext (str):        File extension.
    
    Returns:
      (str):  A temporary file name.
    
    Note:
      - microsecond accuracy should suffice; two subsequent calls on my laptop take ~15-20us.
    """
    
    import datetime as dt
    
    if dir_name is None:  dir_name = homedir()
    
    return dt.datetime.now().strftime(dir_name+'/'+base_name+'_%Y%m%d-%H%M%S.%f.'+ext)  # Date and time with microseconds
    

def tail_file(in_file, out_file, num_lines):
    """Use the tail command in a shell to quickly save the last N lines of a file in a new file.
    
    Parameters:
      in_file (str):    Name of the input file.
      out_file (str):   Name of the output file.
      num_lines (int):  Number of lines to save.
    """
    
    import subprocess
    subprocess.run(['tail -'+str(num_lines)+' '+in_file+' > '+out_file], stdout=subprocess.PIPE,
                   shell=True, check=True).stdout.decode('utf-8')
    
    return


def string_in_file(infile, string):
    """Use the grep command in a shell to quickly check whether a file contains a string.
    
    Parameters:
      infile (str):  Name of the file.
      string (str):  Match string.
    
    Note: fast small files (<~1Mb) and/or few calls (<~100kb).
    """
    
    with open(infile) as f:
        if string in f.read():
            return True
        else:
            return False
    
    return None


def string_in_file_grep(infile, string):
    """Use the grep command in a shell to quickly check whether a file contains a string.
    
    Parameters:
      infile (str):  Name of the file.
      string (str):  Match string.
    
    Note: fast for large files (>~1Mb) and/or multiple calls (>~100kb).
    """
    
    import subprocess
    try:
        subprocess.run(['grep -c '+string+' '+infile], stdout=subprocess.PIPE,
                       shell=True, check=True).stdout.decode('utf-8')
        # Check=false: do not raise an error if string is not found (but return 0)
        result = True  # result = int(count) > 0
        
    except subprocess.CalledProcessError as e:
        result = None
        if e.returncode == 1:  # Exit status: 0: line found, 1: no line found, 2: error
            # count = e.output.decode('utf-8')
            # result = int(count) > 0
            result = False
        else:
            import sluyspy.cli as scli
            scli.warn('string_in_file(): an error occurred when calling grep')
            result = None
    
    return result





# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""AirVisual(Pro) functions for the sluyspy package"""





# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Machine-learning functions for the sluyspy package"""


def stats_from_conf_mat(df):
    """Add common ML statistics to a Pandas DataFrame containing the confusion-matrix elements.
    
    Parameters:
      df (pd.df):  Pandas DataFrame containing the confusion-matrix elements in the following columns:
    
        - tp:  Number or fraction of true positives.
        - tn:  Number or fraction of true negatives.
        - fp:  Number or fraction of false positives.
        - fn:  Number or fraction of false negatives.
    
    Returns:
      (pd.df):  Pandas DataFrame containing different machine-learning statistics, with new columns:
    
        - p:  Total number of actual positive cases in the data.
        - n:  Total number of actual negaitive cases in the data.
    
        - tpr:  TPR: true positive rate (sensitivity): TPR = TP/P.
        - tnr:  TNR: true negative rate (specificity): TNR = TN/N.
        - fpr:  FPR: false positive rate (fall out): FPR = FP/N.
        - fnr:  FNR: false negative rate (miss rate): FNR = FN/P.
    
        - ppv:  PPV - positive predictive value - precision: PPV = TP/(TP+FP).
        - npv:  NPV - negative predictive value: NPV = TN/(TN+FN).
        - fdr:  FDR - false discovery rate: FDR = FP/(TP+FP).
        - for:  FOR - false omission rate: FOR = FN/(TN+FN).
    
        - acc:  accuracy: fraction of true predictions: ACC = (TP+TN)/(P+N).
        - f1:  F1 score: harmonic mean of PPV and TPR: F1 = PPV*TPR/(PPV+TPR).
    """
    
    tp = df.tp
    tn = df.tn
    fp = df.fp
    fn = df.fn
    
    df['p']   = fn + tp                  # Total number of actual positive cases in the data
    df['n']   = fp + tn                  # Total number of actual negaitive cases in the data
    
    df['tpr'] = tp/(tp+fn)               # TPR: true positive rate (sensitivity): TPR = TP/P
    df['tnr'] = tn/(tn+fp)               # TNR: true negative rate (specificity): TNR = TN/N
    
    df['fpr'] = fp/(fp+tn)               # FPR: false positive rate (fall out): FPR = FP/N
    df['fnr'] = fn/(fn+tp)               # FNR: false negative rate (miss rate): FNR = FN/P
    
    df['ppv'] = tp/(tp+fp)               # PPV - positive predictive value - precision: PPV = TP/(TP+FP)
    df['npv'] = tn/(tn+fn)               # NPV - negative predictive value: NPV = TN/(TN+FN)
    
    df['fdr'] = fp/(tp+fp)               # FDR - false discovery rate: FDR = FP/(TP+FP)
    df['for'] = fn/(tn+fn)               # FOR - false omission rate: FOR = FN/(TN+FN)
    
    df['acc'] = (tp+tn)/(tp+tn+fp+fn)    # accuracy: fraction of true predictions: ACC = (TP+TN)/(P+N)
    df['f1']  = (2 * tp)/(2*tp + fp+fn)  # F1 score: harmonic mean of PPV and TPR: F1 = PPV*TPR/(PPV+TPR)
    
    return df


